/*
   polys.p

   Ian Wright 9.5.95

   Routines for collision detection of polygons defined
   by their vertices.

*/

uses rc_graphic;

vars procedure collision_detect get_distance return_intersections
               get_bounded_intersection in_box get_intersection
               make_polygon generate_equations get_line;

/*
         CONTENTS - (Use <ENTER> gg to access required sections)

 define :class Line;
 define :class poly;
 define :method collision_detect( pos1, pos2, polygon:poly ) -> hit -> newpos;
 define procedure get_distance( pos1, pos2 ) -> dist;
 define :method return_intersections( pos1, pos2, polygon:poly ) -> int_points;
 define :method get_bounded_intersection( line1:Line, line2:Line, line1_bounds,
 define procedure in_box(bounds, the_point) -> bool;
 define :method get_intersection( line1:Line, line2:Line ) -> intersect_point;
 define procedure make_polygon(vertices) -> polygon;
 define :method generate_equations( polygon:poly );
 define :method get_line( pos1, pos2 ) -> line_eq;

*/

define :class Line;

   slot Line_type; ;;; y (i.e., y = c) x (i.e., x = c) or n (y = mx + c)
   slot Line_grad;
   slot Line_cnst;

enddefine;

define :class poly;

   slot poly_points = [];
   slot poly_equations = [];

enddefine;

;;; COLLISION_DETECT
;;; Given two points (original point and point to move to) and
;;; a polygon, return true if collision occurs (false otherwise)
;;; and new point.
define :method collision_detect( pos1, pos2, polygon:poly ) -> hit -> newpos;

lvars pos1, pos2, polygon, hit, newpos, intersect_points,
      nearest_point, smallest_dist, dist, a_point;

   ;;; get all bounded intersection points
   return_intersections(pos1, pos2, polygon) -> intersect_points;
   unless null(intersect_points) then
      ;;; find the nearest point
      hd(intersect_points) -> nearest_point;
      get_distance(pos1, nearest_point) -> smallest_dist;
      for a_point in intersect_points do
         get_distance(pos1, a_point) -> dist;
         if dist < smallest_dist then
            a_point -> nearest_point;
            dist -> smallest_dist;
         endif;
      endfor;
      true -> hit;
      nearest_point -> newpos;
   else
      false -> hit;
      pos2 -> newpos;
   endunless;

enddefine;

;;; GET_DISTANCE
;;; Returns distance between two points
define procedure get_distance( pos1, pos2 ) -> dist;

lvars pos1, pos2, dist;

   (pos1(1) - pos2(1)) ** 2 + (pos1(2) - pos2(2)) ** 2 -> dist;
   sqrt(dist) -> dist;

enddefine;

;;; RETURN_INTERSECTIONS
;;; Given two points and a polygon returns all the intersection points
;;; between a line draw between the two points and the sides of the polygon.
define :method return_intersections( pos1, pos2, polygon:poly ) -> int_points;

lvars pos1, pos2, polygon, int_points, movement_eq, count,
      intersect_point;

   get_line(pos1, pos2) -> movement_eq;
   [%
      for count from 1 to length(polygon.poly_equations) do
         unless count = length(polygon.poly_equations) then
            get_bounded_intersection(movement_eq, (polygon.poly_equations)(count),
               [^pos1 ^pos2], [%(polygon.poly_points)(count)%
                  %(polygon.poly_points)(count + 1)%]) -> intersect_point;
         else
            get_bounded_intersection(movement_eq, (polygon.poly_equations)(count),
               [^pos1 ^pos2], [%(polygon.poly_points)(count)%
                  %(polygon.poly_points)(1)%]) -> intersect_point;
         endunless;
         unless intersect_point = false then
            intersect_point
         endunless;
      endfor;
   %] -> int_points;

enddefine;

;;; GET_BOUNDED_INTERSECTION
;;; Given two bounded line equations returns intersect point (if no point
;;; returns false)
define :method get_bounded_intersection( line1:Line, line2:Line, line1_bounds,
   line2_bounds ) -> intersect_point;

lvars line1, line2, line1_bounds, line2_bounds, intersect_point;

   get_intersection(line1, line2) -> intersect_point;
   ;;; intersect_point must be in both bounding boxes
   unless null(intersect_point) then
      unless in_box(line1_bounds, intersect_point) and
            in_box(line2_bounds, intersect_point) then
         false -> intersect_point;
      endunless;
   else
      false -> intersect_point;
   endunless;

enddefine;

;;; IN_BOX
;;; Given two coordinate bounds and a point returns true if point
;;; is within the bounding box defined by the coordinates.
define procedure in_box(bounds, the_point) -> bool;

lvars bounds, the_point, ok, bool, pointA, pointB;

   bounds(1) -> pointA;
   bounds(2) -> pointB;
   false -> ok;
   if pointA(1) > pointB(1) then
      if the_point(1) >= pointB(1) and
            the_point(1) <= pointA(1) then
         true -> ok;
      endif;
   elseif pointA(1) < pointB(1) then
      if the_point(1) >= pointA(1) and
            the_point(1) <= pointB(1) then
         true -> ok;
      endif;
   else
      if the_point(1) = pointA(1) then
         true -> ok;
      endif;
   endif;
   if ok then
      false -> ok;
      if pointA(2) > pointB(2) then
         if the_point(2) >= pointB(2) and
               the_point(2) <= pointA(2) then
            true -> ok;
         endif;
      elseif pointA(2) < pointB(2) then
         if the_point(2) >= pointA(2) and
               the_point(2) <= pointB(2) then
            true -> ok;
         endif;
      else
         if the_point(2) = pointA(2) then
            true -> ok;
         endif;
      endif;
   endif;
   ok -> bool;

enddefine;

;;; GET_INTERSECTION
;;; Given two line equations returns intersect point (if any)
define :method get_intersection( line1:Line, line2:Line ) -> intersect_point;

lvars line1, line2, line3, intersect_point;

   if line1.Line_type = 'n' and line2.Line_type /= 'n' then
      copy(line1) -> line3;
      copy(line2) -> line1;
      line3 -> line2;
   endif;
   [] -> intersect_point;
   if line1.Line_type = 'n' and line2.Line_type = 'n' then
      ;;; two normal line equations
      if line1.Line_grad = line2.Line_grad then
         ;;; leave intersect_point null
      else
         [ ^((line2.Line_cnst - line1.Line_cnst) / (line1.Line_grad
             - line2.Line_grad))
           ^((line2.Line_grad * line1.Line_cnst
             - line1.Line_grad * line2.Line_cnst) / (line2.Line_grad
             - line1.Line_grad))
         ] -> intersect_point;
      endif;
   elseif line1.Line_type = 'x' then
      ;;; line1 is x = c
      switchon line2.Line_type
         case = 'n' then
            [ ^(line1.Line_cnst)
              ^(line1.Line_cnst * line2.Line_grad + line2.Line_cnst)
            ] -> intersect_point;
         case = 'x' then
            ;;; leave intersect_point null
         case = 'y' then
            [ ^(line1.Line_cnst)
              ^(line2.Line_cnst)
            ] -> intersect_point;
      endswitchon;
   elseif line1.Line_type = 'y' then
      ;;; line1 is y = c
      switchon line2.Line_type
         case = 'n' then
            [ ^((line1.Line_cnst - line2.Line_cnst) / line2.Line_grad)
              ^(line1.Line_cnst)
            ] -> intersect_point;
         case = 'x' then
            [ ^(line2.Line_cnst)
              ^(line1.Line_cnst)
            ] -> intersect_point;
         case = 'y' then
            ;;; leave intersect_point null
      endswitchon;
   endif;

enddefine;

;;; MAKE_POLYGON
;;; Given a list of vertices, constructs a polygon.
define procedure make_polygon(vertices) -> polygon;

lvars vertices, polygon, a_point;

   newpoly() -> polygon;
   [%
      for a_point in vertices do
         [ %hd(a_point)% %last(a_point)% ];
      endfor;
   %] -> polygon.poly_points;
   generate_equations(polygon);

enddefine;

/*
    make_polygon([[10 20] [100 40] [30 200]]) =>
*/

;;; GENERATE_EQUATIONS
;;; Given a polygon with vertices defined generates the
;;; line equations corresponding to its sides.
define :method generate_equations( polygon:poly );

lvars polygon, num_points, point_num;

   length(polygon.poly_points) -> num_points;
   [%
      for point_num from 1 to num_points do
         unless point_num = num_points then
            get_line((polygon.poly_points)(point_num),
               (polygon.poly_points)(point_num + 1))
         else
            get_line((polygon.poly_points)(point_num), (polygon.poly_points)(1))
         endunless;
      endfor;
   %] -> polygon.poly_equations;

enddefine;

;;; GET_LINE
;;; Given two points returns the line equation running through them.
define :method get_line( pos1, pos2 ) -> line_eq;

lvars pos1, pos2, line_eq, numer, denom;

   newLine() -> line_eq;
   pos2(1) - pos1(1) -> denom;
   pos2(2) - pos1(2) -> numer;
   if denom = 0 then
      'x' -> line_eq.Line_type;
      pos1(1) -> line_eq.Line_cnst;
   elseif numer = 0 then
      'y' -> line_eq.Line_type;
      pos1(2) -> line_eq.Line_cnst;
   else
      'n' -> line_eq.Line_type;
      numer / denom -> line_eq.Line_grad;
      pos1(2) - (line_eq.Line_grad * pos1(1)) -> line_eq.Line_cnst;
   endif;

enddefine;
